#include <iostream>
#include <vector>
#include "NAOKinematics.h"
#include "KMat.hpp"
//
#include <fstream>
#include <sstream>
#include <string>
#include <typeinfo>
#include <stdlib.h>
#include <iomanip>
//

using namespace std;
using namespace KMath::KMat;
using namespace KDeviceLists;

void multiplyMatrices(double A[3][3], double B[3][3], double C[3][3]);

int main() {
	NAOKinematics nkin;
	NAOKinematics::kmatTable output1, output2, output3, output4, output5, T;

	std::vector<float> joints(NUMOFJOINTS);
	double pi = KMath::KMat::transformations::PI;
	//Left Hand
	joints[L_ARM + SHOULDER_PITCH] = 0;																		//0.2
	joints[L_ARM + SHOULDER_ROLL] = 1.3;																//0.1
	joints[L_ARM + ELBOW_YAW] = 0.3;																		//0
	joints[L_ARM + ELBOW_ROLL] = 0.1;																					//0
	joints[L_ARM + WRIST_YAW] = 0;																							//0
	//Right Hand
	joints[R_ARM + SHOULDER_PITCH] = M_PI_2;
	joints[R_ARM + SHOULDER_ROLL] = -M_PI_4;
	joints[R_ARM + ELBOW_YAW] = 0;
	joints[R_ARM + ELBOW_ROLL] = 0;
	joints[R_ARM + WRIST_YAW] = M_PI / 3535;
	//Left Leg
	joints[L_LEG + HIP_YAW_PITCH] = 0;
	joints[L_LEG + HIP_ROLL] = 0;
	joints[L_LEG + HIP_PITCH] = 0;
	joints[L_LEG + KNEE_PITCH] = M_PI / 4;
	joints[L_LEG + ANKLE_PITCH] = -M_PI / 4;
	joints[L_LEG + ANKLE_ROLL] = 0;
	//Right Leg
	joints[R_LEG + HIP_YAW_PITCH] = 0;
	joints[R_LEG + HIP_ROLL] = 0;
	joints[R_LEG + HIP_PITCH] = 0;
	joints[R_LEG + KNEE_PITCH] = 0;
	joints[R_LEG + ANKLE_PITCH] = 0;
	joints[R_LEG + ANKLE_ROLL] = 0;
	//Head
	joints[HEAD + YAW] = M_PI_2;
	joints[HEAD + PITCH] = 0;

	//Now we can set the joints to the class
	nkin.setJoints(joints);

	output1 = nkin.getForwardEffector((NAOKinematics::Effectors)CHAIN_L_ARM);

	//Right Hand
	output2 = nkin.getForwardEffector((NAOKinematics::Effectors)CHAIN_R_ARM);

	//Left Leg
	output3 = nkin.getForwardEffector((NAOKinematics::Effectors)CHAIN_L_LEG);

	//Right Leg
	output4 = nkin.getForwardEffector((NAOKinematics::Effectors)CHAIN_R_LEG);

	//Camera
	output5 = nkin.getForwardEffector(NAOKinematics::EFF_CAMERA_BOT);

	std::cout << "x = " << output1(0, 3) << " y = " << output1(1, 3) << " z = " << output1(2, 3) << std::endl;
	std::cout << "x = " << output2(0, 3) << " y = " << output2(1, 3) << " z = " << output2(2, 3) << std::endl;
	std::cout << "x = " << output3(0, 3) << " y = " << output3(1, 3) << " z = " << output3(2, 3) << std::endl;
	std::cout << "x = " << output4(0, 3) << " y = " << output4(1, 3) << " z = " << output4(2, 3) << std::endl;
	std::cout << "x = " << output5(0, 3) << " y = " << output5(1, 3) << " z = " << output5(2, 3) << std::endl;

	std::cout << output1(0, 0) << " " << output1(0, 1) << " " << output1(0, 2) << " " << output1(0, 3) << std::endl;
	std::cout << output1(1, 0) << " " << output1(1, 1) << " " << output1(1, 2) << " " << output1(1, 3) << std::endl;
	std::cout << output1(2, 0) << " " << output1(2, 1) << " " << output1(2, 2) << " " << output1(2, 3) << std::endl;


	//MYPART
	//
	ifstream myfile("deviceinit.txt");
	string initposline;
	string initquatline;
	string initline;
	vector<string> initquattokens;
	vector<string> initpostokens;
	if (myfile.is_open()) {
		for (int i = 0; i < 7; i++) {
			getline(myfile, initline);
			if (i == 5) {
				initposline = initline;
			}
			if (i == 6) {
				initquatline = initline;
			}
		}


		cout << "Initial posline: " << initposline << '\n';
		cout << "Initial quatline: " << initquatline << '\n';

		//Here we extract the values for the four quaternions.
		string initpartial;
		istringstream initissq(initquatline);
		string inittoken;
		while (getline(initissq, inittoken, '\t'))
			initquattokens.push_back(inittoken);

		for (int i = 0; i < initquattokens.size(); i++) {
			cout << initquattokens[i] << " ";
		}

		//Here we extract the values for the three positions.
		//string partial;
		istringstream initissp(initposline);
		//string token;
		while (getline(initissp, inittoken, '\t'))
			initpostokens.push_back(inittoken);

		for (int i = 0; i < initpostokens.size(); i++) {
			cout << initpostokens[i] << " ";
		}


		myfile.close();
	}
	else cout << "Uanble to open file \n";

	//double initx = atof(initpostokens[1].c_str()) ; 
	//double inity = atof(initpostokens[2].c_str()); 
	//double initz = atof(initpostokens[3].c_str()) ; 

	double initx = -atof(initpostokens[3].c_str())/4; //NAO x = STEM -z
	double inity = -atof(initpostokens[1].c_str())/4; //NAO y = STEM -x
	double initz = atof(initpostokens[2].c_str())/4; //NAO z = STEM y

	cout << "Initial xyz for STEM: x = " << initx << " y = " << inity << " z = " << initz << endl;
	//This is for left hand only, for now
	//double naoinitx = 110.010564327;
	//double naoinity = 127.078607678;
	//double naoinitz = -55.2914999425;

	//Pose init
	//double naoinitx = 110;
	//double naoinity = 127;
	//double naoinitz = -55.3;

	//double naoinitx = 218.52613985538483;
	//double naoinity = 110.9393984079361;
	//double naoinitz = 87.69000321626663;

	//Pose zero
	double naoinitx = 219;
	double naoinity = 111;
	double naoinitz = 87.7;

	//Pose init
	//double naoinitimatrix[3][3] = { { 0.874904990196228, -0.4527558386325836, 0.17191152274608612 },{ -0.0094371372833848, 0.33896711468696594, 0.9407508969306946 },{ -0.4842028021812439, -0.8246899843215942, 0.2922913134098053 } };
	//Pose zero
	double naoinitimatrix[3][3] = { { 0.9996573328971863,0.026176951825618744 , 2.802596928649634e-45 },{ -0.026176951825618744 , 0.9996573328971863,5.605193857299268e-45 },{ -2.802596928649634e-45 ,-5.605193857299268e-45 ,1.0 } };


	//Should do whole process in C++, but this is from Matlab. ex = -1.23194408417 ey = 0.5057477355 ez = -0.0120304897428 THIS IS THOROUGHLY TESTED 
	//double naoinitimatrix[3][3] = { {0.8747,-0.4529,0.1724},{-0.0105,0.3379,0.9411},{-0.4845,-0.8251,0.2908} };
	//Har dobbeltsjekket med MATlab 23.05.: Dette er riktig rekkefolge p quaternionene!!!!
	double initqw = atof(initquattokens[1].c_str());
	double initqx = atof(initquattokens[4].c_str());
	double initqy = atof(initquattokens[3].c_str());
	double initqz = atof(initquattokens[2].c_str());

	cout << "qx = " << initqx << "  qz = " << initqy << "  qz = " << initqz << "  qw = " << initqw << endl;
	//Normalizing the initial quaternions
	double initqsum = sqrt(initqw*initqw + initqx*initqx + initqy*initqy + initqz*initqz);
	initqw = initqw / initqsum;
	initqx = initqx / initqsum;
	initqy = initqy / initqsum;
	initqz = initqz / initqsum;

	cout << "New initqsum is: " << sqrt(initqw*initqw + initqx*initqx + initqy*initqy + initqz*initqz) << endl;
	
	//stackoverflow.com/questions/11667783/quaternion-and-normalization
	
	double steminitmatrix[3][3] = { { 0,0,0 },{ 0,0,0 },{ 0,0,0 } };
	/*
	//Here it is rotated to match NAO's coordiante system, post-multiplied
	steminitmatrix[0][0] = -(2.0f*initqx*initqz + 2.0f*initqy*initqw);
	steminitmatrix[0][1] = -(1.0f - 2.0f*initqy*initqy - 2.0f*initqz*initqz);
	steminitmatrix[0][2] = 2.0f*initqx*initqy - 2.0f*initqz*initqw;
	steminitmatrix[1][0] = -(2.0f*initqy*initqz - 2.0f*initqx*initqw);
	steminitmatrix[1][1] = -(2.0f*initqx*initqy + 2.0f*initqz*initqw);
	steminitmatrix[1][2] = 1.0f - 2.0f*initqx*initqx - 2.0f*initqz*initqz;
	steminitmatrix[2][0] = -(1.0f - 2.0f*initqx*initqx - 2.0f*initqy*initqy);
	steminitmatrix[2][1] = -(2.0f*initqx*initqz - 2.0f*initqy*initqw);
	steminitmatrix[2][2] = 2.0f*initqy*initqz + 2.0f*initqx*initqw;
	
	*/
	
	steminitmatrix[0][0] = 1.0f - 2.0f*initqy*initqy - 2.0f*initqz*initqz;
	steminitmatrix[0][1] = 2.0f*initqx*initqy - 2.0f*initqz*initqw;
	steminitmatrix[0][2] = 2.0f*initqx*initqz + 2.0f*initqy*initqw;
	steminitmatrix[1][0] = 2.0f*initqx*initqy + 2.0f*initqz*initqw;
	steminitmatrix[1][1] = 1.0f - 2.0f*initqx*initqx - 2.0f*initqz*initqz;
	steminitmatrix[1][2] = 2.0f*initqy*initqz - 2.0f*initqx*initqw;
	steminitmatrix[2][0] = 2.0f*initqx*initqz - 2.0f*initqy*initqw;
	steminitmatrix[2][1] = 2.0f*initqy*initqz + 2.0f*initqx*initqw;
	steminitmatrix[2][2] = 1.0f - 2.0f*initqx*initqx - 2.0f*initqy*initqy;
	/*
	//THIS IS TESTING TRANSPOSE
	steminitmatrix[0][0] = 1.0f - 2.0f*initqy*initqy - 2.0f*initqz*initqz;
	steminitmatrix[0][1] = 2.0f*initqx*initqy + 2.0f*initqz*initqw;
	steminitmatrix[0][2] = 2.0f*initqx*initqz - 2.0f*initqy*initqw;
	steminitmatrix[1][0] = 2.0f*initqx*initqy - 2.0f*initqz*initqw;
	steminitmatrix[1][1] = 1.0f - 2.0f*initqx*initqx - 2.0f*initqz*initqz;
	steminitmatrix[1][2] = 2.0f*initqy*initqz + 2.0f*initqx*initqw;
	steminitmatrix[2][0] = 2.0f*initqx*initqz + 2.0f*initqy*initqw;
	steminitmatrix[2][1] = 2.0f*initqy*initqz - 2.0f*initqx*initqw;
	steminitmatrix[2][2] = 1.0f - 2.0f*initqx*initqx - 2.0f*initqy*initqy;
	*/

	cout << "STEMInitmatrix: " << endl;
	cout << steminitmatrix[0][0] << " " << steminitmatrix[0][1] << "  " << steminitmatrix[0][2] << endl;
	cout << steminitmatrix[1][0] << " " << steminitmatrix[1][1] << "  " << steminitmatrix[1][2] << endl;
	cout << steminitmatrix[2][0] << " " << steminitmatrix[2][1] << "  " << steminitmatrix[2][2] << endl;
	//
	ifstream mynewfile("device3.txt");
	string posline;
	string quatline;
	string line;
	vector<string> quattokens;
	vector<string> postokens;
	if (mynewfile.is_open()) {
		for (int i = 0; i < 7; i++) {
			getline(mynewfile, line);
			if (i == 5) {
				posline = line;
			}
			if (i == 6) {
				quatline = line;
			}
		}

		cout << "Posline: " << posline << '\n';
		cout << "Quatline: " << quatline << '\n';

		//Here we extract the values for the four quaternions.
		string partial;
		istringstream issq(quatline);
		string token;
		while (getline(issq, token, '\t'))
			quattokens.push_back(token);

		for (int i = 0; i < quattokens.size(); i++) {
			cout << quattokens[i] << " ";
		}
		//Here we extract the values for the three positions.
		//string partial;
		istringstream issp(posline);
		//string token;
		while (getline(issp, token, '\t'))
			postokens.push_back(token);

		for (int i = 0; i < postokens.size(); i++) {
			cout << postokens[i] << " ";
		}


		mynewfile.close();
	}
	else cout << "Uanble to open file \n";


	//double x = atof(postokens[1].c_str()) ; 
	//double y = atof(postokens[2].c_str()) ;
	//double z = atof(postokens[3].c_str()) ; 

	double x = -atof(postokens[3].c_str())/4; //NAO x = STEM -z
	double y = -atof(postokens[1].c_str())/4; //NAO y = STEM -x
	double z = atof(postokens[2].c_str())/4; //NAO z = STEM y

	cout << "New xyz for STEM: x = " << x << " y = " << y << " z = " << z << endl;
	//Har dobbeltsjekket med MATlab 23.05.: Dette er riktig rekkefolge p quaternionene!!!!
	double qw = atof(quattokens[1].c_str());
	double qx = atof(quattokens[4].c_str());
	double qy = atof(quattokens[3].c_str());
	double qz = atof(quattokens[2].c_str());

	cout << "qx = " << qx << "  qy = " << qy << "  qz = " << qz << "  qw = " << qw << endl;

	double qsum = sqrt(qw*qw + qx*qx + qy*qy + qz*qz);
	qw = qw / qsum;
	qx = qx / qsum;
	qy = qy / qsum;
	qz = qz / qsum;
		
	double stemmatrix[3][3] = { { 0,0,0 },{ 0,0,0 },{ 0,0,0 } };
	/*
	//This rotation should be correct, 23.05, post-multiplied
	stemmatrix[0][0] = -(2.0f*qx*qz + 2.0f*qy*qw);
	stemmatrix[0][1] = -(1.0f - 2.0f*qy*qy - 2.0f*qz*qz);
	stemmatrix[0][2] = 2.0f*qx*qy - 2.0f*qz*qw;
	stemmatrix[1][0] = -(2.0f*qy*qz - 2.0f*qx*qw);
	stemmatrix[1][1] = -(2.0f*qx*qy + 2.0f*qz*qw);
	stemmatrix[1][2] = 1.0f - 2.0f*qx*qx - 2.0f*qz*qz;
	stemmatrix[2][0] = -(1.0f - 2.0f*qx*qx - 2.0f*qy*qy);
	stemmatrix[2][1] = -(2.0f*qx*qz - 2.0f*qy*qw);
	stemmatrix[2][2] = 2.0f*qy*qz + 2.0f*qx*qw;
	*/
	stemmatrix[0][0] = 1.0f - 2.0f*qy*qy - 2.0f*qz*qz;
	stemmatrix[0][1] = 2.0f*qx*qy - 2.0f*qz*qw;
	stemmatrix[0][2] = 2.0f*qx*qz + 2.0f*qy*qw;
	stemmatrix[1][0] = 2.0f*qx*qy + 2.0f*qz*qw;
	stemmatrix[1][1] = 1.0f - 2.0f*qx*qx - 2.0f*qz*qz;
	stemmatrix[1][2] = 2.0f*qy*qz - 2.0f*qx*qw;
	stemmatrix[2][0] = 2.0f*qx*qz - 2.0f*qy*qw;
	stemmatrix[2][1] = 2.0f*qy*qz + 2.0f*qx*qw;
	stemmatrix[2][2] = 1.0f - 2.0f*qx*qx - 2.0f*qy*qy;
	/*
	//THIS IS TESTING TRANSPOSE
	stemmatrix[0][0] = 1.0f - 2.0f*qy*qy - 2.0f*qz*qz;
	stemmatrix[0][1] = 2.0f*qx*qy + 2.0f*qz*qw;
	stemmatrix[0][2] = 2.0f*qx*qz - 2.0f*qy*qw;
	stemmatrix[1][0] = 2.0f*qx*qy - 2.0f*qz*qw;
	stemmatrix[1][1] = 1.0f - 2.0f*qx*qx - 2.0f*qz*qz;
	stemmatrix[1][2] = 2.0f*qy*qz + 2.0f*qx*qw;
	stemmatrix[2][0] = 2.0f*qx*qz + 2.0f*qy*qw;
	stemmatrix[2][1] = 2.0f*qy*qz - 2.0f*qx*qw;
	stemmatrix[2][2] = 1.0f - 2.0f*qx*qx - 2.0f*qy*qy;
	*/

	cout << "New STEMmatrix: " << endl;
	cout << stemmatrix[0][0] << " " << stemmatrix[0][1] << "  " << stemmatrix[0][2] << endl;
	cout << stemmatrix[1][0] << " " << stemmatrix[1][1] << "  " << stemmatrix[1][2] << endl;
	cout << stemmatrix[2][0] << " " << stemmatrix[2][1] << "  " << stemmatrix[2][2] << endl;

	//Dette er dobbeltsjekket i MATlab, bade determinant og 1/determinant stemmer helt!
	double initdet = steminitmatrix[0][0] * (steminitmatrix[1][1] * steminitmatrix[2][2] - steminitmatrix[2][1] * steminitmatrix[1][2]) -
		steminitmatrix[0][1] * (steminitmatrix[1][0] * steminitmatrix[2][2] - steminitmatrix[1][2] * steminitmatrix[2][0]) +
		steminitmatrix[0][2] * (steminitmatrix[1][0] * steminitmatrix[2][1] - steminitmatrix[1][1] * steminitmatrix[2][0]);


	double invinitdet = 1 / initdet;
	cout << "This is the determinant:" << initdet <<", and this is the inverted determinant: " << invinitdet << endl;

	double steminitmatrixinv[3][3] = { {0,0,0},{0,0,0},{0,0,0} }; // inverse of matrix m
	//This is also tested in MATLAB, should be correct per 23.05.
	steminitmatrixinv[0][0] = (steminitmatrix[1][1] * steminitmatrix[2][2] - steminitmatrix[2][1] * steminitmatrix[1][2]) * invinitdet;
	steminitmatrixinv[0][1] = (steminitmatrix[0][2] * steminitmatrix[2][1] - steminitmatrix[0][1] * steminitmatrix[2][2]) * invinitdet;
	steminitmatrixinv[0][2] = (steminitmatrix[0][1] * steminitmatrix[1][2] - steminitmatrix[0][2] * steminitmatrix[1][1]) * invinitdet;
	steminitmatrixinv[1][0] = (steminitmatrix[1][2] * steminitmatrix[2][0] - steminitmatrix[1][0] * steminitmatrix[2][2]) * invinitdet;
	steminitmatrixinv[1][1] = (steminitmatrix[0][0] * steminitmatrix[2][2] - steminitmatrix[0][2] * steminitmatrix[2][0]) * invinitdet;
	steminitmatrixinv[1][2] = (steminitmatrix[1][0] * steminitmatrix[0][2] - steminitmatrix[0][0] * steminitmatrix[1][2]) * invinitdet;
	steminitmatrixinv[2][0] = (steminitmatrix[1][0] * steminitmatrix[2][1] - steminitmatrix[2][0] * steminitmatrix[1][1]) * invinitdet;
	steminitmatrixinv[2][1] = (steminitmatrix[2][0] * steminitmatrix[0][1] - steminitmatrix[0][0] * steminitmatrix[2][1]) * invinitdet;
	steminitmatrixinv[2][2] = (steminitmatrix[0][0] * steminitmatrix[1][1] - steminitmatrix[1][0] * steminitmatrix[0][1]) * invinitdet;

	//Test of 

	/*https://msdn.microsoft.com/en-us/library/hh873134.aspx
	
	*/
	double realmatrix[3][3] = {{0,0,0}, {0,0,0} ,{0,0,0}};
	cout << "Here is the actual rotation of the STEM-tracker: " << endl;
	for (int row = 0; row < 3; row++) {
		cout << '\n';
		for (int col = 0; col < 3; col++) {
			// Multiply the row of initmatrixinv by the column of matrix to get the row, column of product.
			realmatrix[row][col] = 0;//
			for (int inner = 0; inner < 3; inner++) {
				realmatrix[row][col] += steminitmatrixinv[row][inner] * stemmatrix[inner][col];
			}
			std::cout << realmatrix[row][col] << "  ";
		}
		std::cout << "\n";
	}
	cout << "\n";



	double det = naoinitimatrix[0][0] * (naoinitimatrix[1][1] * naoinitimatrix[2][2] - naoinitimatrix[2][1] * naoinitimatrix[1][2]) -
		naoinitimatrix[0][1] * (naoinitimatrix[1][0] * naoinitimatrix[2][2] - naoinitimatrix[1][2] * naoinitimatrix[2][0]) +
		naoinitimatrix[0][2] * (naoinitimatrix[1][0] * naoinitimatrix[2][1] - naoinitimatrix[1][1] * naoinitimatrix[2][0]);



	cout << "The determinant of NAO initalt matrix: " << setprecision(20) << det << endl;
	cout << '\n' << endl;


	
	double transpose[3][3] = { {0,0,0},{0,0,0},{0,0,0} };
	for (int row = 0; row < 3; row++) {
		for (int col = 0; col < 3; col++) {
			transpose[row][col] = naoinitimatrix[col][row];
		}
	}
		
	double AtA[3][3] = { { 0,0,0 },{ 0,0,0 },{ 0,0,0 } };
	multiplyMatrices(transpose,naoinitimatrix,AtA);
	


	cout << "If NAO's initial rotation matrix is orthogonal, this matrix should equal I: " << endl;
	cout << "\n" << endl;
	cout << AtA[0][0] << " " << AtA[0][1] << "  " << AtA[0][2] << endl;
	cout << AtA[1][0] << " " << AtA[1][1] << "  " << AtA[1][2] << endl;
	cout << AtA[2][0] << " " << AtA[2][1] << "  " << AtA[2][2] << endl;


	ofstream matrixfile("stemrotation.txt");//

	for (int j = 0; j<3; j++) {
		for (int i = 0; i < 3; i++) {
			matrixfile << realmatrix[j][i] << '\n';//
		}
	}
	cout << endl;
	matrixfile.close();//

	//double rotation[3][3] = { { 0,-1,0 },{ 0,0,1 },{ -1,0,0 } };
	//double rotrealmatrix[3][3] = { { 0,0,0 },{ 0,0,0 } ,{ 0,0,0 } };

	//multiplyMatrices(realmatrix, rotation, rotrealmatrix);

	//cout << "This is the rotated real rotation:" << endl;
	//cout << rotrealmatrix[0][0] << " " << rotrealmatrix[0][1] << " " << rotrealmatrix[0][2] << endl;
	//cout << rotrealmatrix[1][0] << " " << rotrealmatrix[1][1] << " " << rotrealmatrix[1][2] << endl;
	//cout << rotrealmatrix[2][0] << " " << rotrealmatrix[2][1] << " " << rotrealmatrix[2][2] << endl;

	
	double finalmatrix[3][3] = { { 0,0,0 },{ 0,0,0 } ,{ 0,0,0 } };
	cout << "Here is new transform matrix for NAO: " << endl;
	for (int row = 0; row < 3; row++) {
		cout << '\n';
		for (int col = 0; col < 3; col++) {
			// Multiply the row of initmatrixinv by the column of matrix to get the row, column of product.
			finalmatrix[row][col] = 0;//
			for (int inner = 0; inner < 3; inner++) {
				finalmatrix[row][col] += naoinitimatrix[row][inner] * realmatrix[inner][col];
			}
			std::cout << finalmatrix[row][col] << "  ";
		}
		std::cout << "\n";
	}

	double detfin = finalmatrix[0][0] * (finalmatrix[1][1] * finalmatrix[2][2] - finalmatrix[2][1] * finalmatrix[1][2]) -
		finalmatrix[0][1] * (finalmatrix[1][0] * finalmatrix[2][2] - finalmatrix[1][2] * finalmatrix[2][0]) +
		finalmatrix[0][2] * (finalmatrix[1][0] * finalmatrix[2][1] - finalmatrix[1][1] * finalmatrix[2][0]);

	cout << "\n" << endl;
	cout << "The determinant of desired orientation for NAO's hand: " << detfin << endl;
	cout << '\n' << endl;

	//Here begins part after correct matrix is achieved. It is checked with Matlab that this is correct
	
	output1(0, 0) = finalmatrix[0][0];
	output1(0, 1) = finalmatrix[0][1];
	output1(0, 2) = finalmatrix[0][2];
	output1(1, 0) = finalmatrix[1][0];
	output1(1, 1) = finalmatrix[1][1];
	output1(1, 2) = finalmatrix[1][2];
	output1(2, 0) = finalmatrix[2][0];
	output1(2, 1) = finalmatrix[2][1];
	output1(2, 2) = finalmatrix[2][2];
	
	output1(0, 3) = naoinitx;
	output1(1, 3) = naoinity;
	output1(2, 3) = naoinitz;
	
	//output1(0, 3) = (x - initx) + naoinitx;
	//output1(1, 3) = (y - inity) + naoinity;
	//output1(2, 3) = (z - initz) + naoinitz;

	//output1(0, 3) = 217.264056;
	//output1(1, 3) = 130.01058996;
	//output1(2, 3) = 94.608575;

	//output1(0, 3) = 219;
	//output1(1, 3) = 111;
	//output1(2, 3) = 88;


	cout << "Start x: " << naoinitx << endl;
	cout << "Start y: " << naoinity << endl;
	cout << "Start z: " << naoinitz << endl;

	cout << "Final x = " << output1(0, 3) << endl;
	cout << "Final y = " << output1(1, 3) << endl;
	cout << "Final z = " << output1(2, 3) << endl;

	cout << "Change in x: " << x - initx << endl;
	cout << "Change in y: " << y - inity << endl;
	cout << "Change in z: " << z - initz << endl;
	
	/*
	output1(0, 3) = round((x - initx + naoinitx) / scale)*scale;
	output1(1, 3) = round((y - inity + naoinity)/scale)*scale;
	output1(2, 3) = round((z - initz + naoinitz) / scale)*scale;
	
	cout << '\n' << endl;
	cout << "NEW, CORRECT TRANSFORMATION MATRIX" << endl;
	cout << output1(0, 0) << " " << output1(0, 1) << "  " << output1(0, 2) << "  " << output1(0, 3) << endl;
	cout << output1(1, 0) << " " << output1(1, 1) << "  " << output1(1, 2) << "  " << output1(1, 3) << endl;
	cout << output1(2, 0) << " " << output1(2, 1) << "  " << output1(2, 2) << "  " << output1(2, 3) << endl;
	cout << '\n' << endl;
	*/
	//MYPARTEND
	
	vector<vector<float> > result;
	result = nkin.inverseLeftHand(output1);
	if(!result.empty()){
		cout << "--Solution exists 1" << endl;
		//MY PART
		ofstream anglefile("deviceangles.txt");//
			
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << endl;
			anglefile << result[0][j] << '\n';//
		}
		cout << endl;
		anglefile.close();//
		//Under construction
		/*joints[L_ARM+SHOULDER_PITCH]=0;
		joints[L_ARM+SHOULDER_ROLL]=0;
		joints[L_ARM+ELBOW_YAW]=0;
		joints[L_ARM+ELBOW_ROLL]=0;
		joints[L_ARM+WRIST_YAW]=0.0;
		nkin.setJoints(joints);
		result = nkin.jacobianInverseLeftHand(output1);
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;*/


	}
	//NEWLY ADDED
	else{
		cout << "No solution for left arm" << std::endl;
		cout << '\n' << endl;
		ofstream anglefile("deviceangles.txt");//
		anglefile << "No valid solution" << '\n';//
		anglefile.close();//
	}
	//END NEWLY ADDED
	
	result = nkin.inverseRightHand(output2);
	if(!result.empty()){
		cout << "--Solution exists 2" << endl;
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;
		//Under construction
		/*
		joints[R_ARM+SHOULDER_PITCH]=0;
		joints[R_ARM+SHOULDER_ROLL]=0;
		joints[R_ARM+ELBOW_YAW]=0;
		joints[R_ARM+ELBOW_ROLL]=0;
		joints[R_ARM+WRIST_YAW]=0.0;
		nkin.setJoints(joints);
		result = nkin.jacobianInverseRightHand(output2);
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;
		*/
	}
	//NEWLY ADDED
	else {
		cout << "No solution for right arm" << std::endl;
	}
	//END NEWLY ADDED


	result = nkin.inverseLeftLeg(output3);
	if(!result.empty()){
		cout << "--Solution exists 3" << endl;
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;
		joints[L_LEG+HIP_YAW_PITCH]=0;
		joints[L_LEG+HIP_ROLL]=0;
		joints[L_LEG+HIP_PITCH]=0;
		joints[L_LEG+KNEE_PITCH]=0;
		joints[L_LEG+ANKLE_PITCH]=0;
		joints[L_LEG+ANKLE_ROLL]=0;
		nkin.setJoints(joints);
		result = nkin.jacobianInverseLeftLeg(output3);
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;
	}
	result = nkin.inverseRightLeg(output4);
	if(!result.empty()){
		cout << "--Solution exists 4" << endl;
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;
		joints[R_LEG+HIP_YAW_PITCH]=0;
		joints[R_LEG+HIP_ROLL]=0;
		joints[R_LEG+HIP_PITCH]=0;
		joints[R_LEG+KNEE_PITCH]=0;
		joints[R_LEG+ANKLE_PITCH]=0;
		joints[R_LEG+ANKLE_ROLL]=0;
		nkin.setJoints(joints);
		result = nkin.jacobianInverseRightLeg(output4);
		for(int j=0; j<result[0].size(); j++){
			cout << "angle" << j << " = " << result[0][j] << " ";
		}
		cout << endl;
	}
	return 0;
}


void multiplyMatrices(double A[3][3], double B[3][3], double C[3][3]) {
	for (int row = 0; row < 3; row++) {
		for (int col = 0; col < 3; col++) {
			C[row][col] = 0;//
			for (int inner = 0; inner < 3; inner++) {
				C[row][col] += A[row][inner] * B[inner][col];
			}
		}
	}
}